import sys

sys.path.append("../")
import NEMtropy.graph_classes as sample
import NEMtropy.models_functions as mof
import NEMtropy.solver_functions as sof
import numpy as np
import unittest  # test tool
import NEMtropy.matrix_generator as mg


class MyTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_quasinewton_0(self):
        n, s = (4, 25)

        A = mg.random_weighted_matrix_generator_dense(
            n, sup_ext=10, sym=False, seed=s, intweights=True
        )

        bA = np.array([[1 if aa != 0 else 0 for aa in a] for a in A])

        k_out = np.sum(bA, axis=1)
        k_in = np.sum(bA, axis=0)
        s_out = np.sum(A, axis=1)
        s_in = np.sum(A, axis=0)

        x0 = 0.9 * np.ones(n * 4)
        args = (k_out, k_in, s_out, s_in)

        fun = lambda x: -mof.loglikelihood_prime_decm_exp(x, args)
        fun_jac = lambda x: -mof.loglikelihood_hessian_diag_decm_exp(
            x, args
        )
        step_fun = lambda x: -mof.loglikelihood_decm_exp(x, args)
        lin_fun = lambda x: mof.linsearch_fun_DECM_exp(x, (mof.loglikelihood_decm_exp, args))
        hes_reg = sof.matrix_regulariser_function

        sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            fun_jac=fun_jac,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-10,
            max_steps=300,
            method="quasinewton",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )
        sol = np.exp(-sol)

        ek = mof.expected_decm(sol)
        k = np.concatenate((k_out, k_in, s_out, s_in))
        err = np.max(np.abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 0: error = {}'.format(err))
        # print('method = {}, matrix {}x{}'.format('quasinewton', n, n))

        # test result
        self.assertTrue(err < 1e-1)

    def test_quasinewton_1(self):
        n, s = (4, 25)

        A = mg.random_weighted_matrix_generator_dense(
            n, sup_ext=10, sym=False, seed=s, intweights=True
        )
        A[0, :] = 0

        bA = np.array([[1 if aa != 0 else 0 for aa in a] for a in A])

        k_out = np.sum(bA, axis=1)
        k_in = np.sum(bA, axis=0)
        s_out = np.sum(A, axis=1)
        s_in = np.sum(A, axis=0)

        x0 = 0.9 * np.ones(n * 4)
        args = (k_out, k_in, s_out, s_in)

        fun = lambda x: -mof.loglikelihood_prime_decm_exp(x, args)
        fun_jac = lambda x: -mof.loglikelihood_hessian_diag_decm_exp(
            x, args
        )
        step_fun = lambda x: -mof.loglikelihood_decm_exp(x, args)
        lin_fun = lambda x: mof.linsearch_fun_DECM_exp(x, (mof.loglikelihood_decm_exp, args))
        hes_reg = sof.matrix_regulariser_function

        sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            fun_jac=fun_jac,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-10,
            max_steps=300,
            method="quasinewton",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )
        sol = np.exp(-sol)

        ek = mof.expected_decm(sol)
        k = np.concatenate((k_out, k_in, s_out, s_in))
        err = np.max(np.abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 0: error = {}'.format(err))
        # print('method = {}, matrix {}x{}'.format('quasinewton', n, n))

        # test result
        self.assertTrue(err < 1e-1)

    @unittest.skip("skip large graph")
    def test_quasinewton_2(self):
        n, s = (40, 25)

        A = mg.random_weighted_matrix_generator_dense(
            n, sup_ext=10, sym=False, seed=s, intweights=True
        )

        bA = np.array([[1 if aa != 0 else 0 for aa in a] for a in A])

        k_out = np.sum(bA, axis=1)
        k_in = np.sum(bA, axis=0)
        s_out = np.sum(A, axis=1)
        s_in = np.sum(A, axis=0)

        x0 = 0.9 * np.ones(n * 4)
        args = (k_out, k_in, s_out, s_in)

        fun = lambda x: -mof.loglikelihood_prime_decm_exp(x, args)
        fun_jac = lambda x: -mof.loglikelihood_hessian_diag_decm_exp(
            x, args
        )
        step_fun = lambda x: -mof.loglikelihood_decm_exp(x, args)
        lin_fun = lambda x: mof.linsearch_fun_DECM_exp(x, (step_fun,))
        hes_reg = sof.matrix_regulariser_function

        sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            fun_jac=fun_jac,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-10,
            max_steps=300,
            method="quasinewton",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )
        sol = np.exp(-sol)

        ek = mof.expected_decm(sol)
        k = np.concatenate((k_out, k_in, s_out, s_in))
        err = np.max(np.abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 0: error = {}'.format(err))
        # print('method = {}, matrix {}x{}'.format('quasinewton', n, n))

        # test result
        self.assertTrue(err < 1e-1)

    @unittest.skip("skip large graph")
    def test_quasinewton_3(self):

        n, s = (40, 25)

        A = mg.random_weighted_matrix_generator_dense(
            n, sup_ext=10, sym=False, seed=s, intweights=True
        )
        A[0, :] = 0

        bA = np.array([[1 if aa != 0 else 0 for aa in a] for a in A])

        k_out = np.sum(bA, axis=1)
        k_in = np.sum(bA, axis=0)
        s_out = np.sum(A, axis=1)
        s_in = np.sum(A, axis=0)

        x0 = 0.9 * np.ones(n * 4)
        args = (k_out, k_in, s_out, s_in)

        fun = lambda x: -mof.loglikelihood_prime_decm_exp(x, args)
        fun_jac = lambda x: -mof.loglikelihood_hessian_diag_decm_exp(
            x, args
        )
        step_fun = lambda x: -mof.loglikelihood_decm_exp(x, args)
        lin_fun = lambda x: mof.linsearch_fun_DECM_exp(x, (step_fun,))
        hes_reg = sof.matrix_regulariser_function

        sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            fun_jac=fun_jac,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-10,
            max_steps=300,
            method="quasinewton",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )
        sol = np.exp(-sol)

        ek = mof.expected_decm(sol)
        k = np.concatenate((k_out, k_in, s_out, s_in))
        err = np.max(np.abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 1: error = {}'.format(err))

        # test result
        self.assertTrue(err < 1e-1)
        # print('method = {}, matrix {}x{} with zeros'.format('quasinewton', n, n))

    def test_newton_0(self):
        # x0 relies heavily on x0
        n, s = (4, 25)

        A = mg.random_weighted_matrix_generator_dense(
            n, sup_ext=10, sym=False, seed=s, intweights=True
        )

        bA = np.array([[1 if aa != 0 else 0 for aa in a] for a in A])

        k_out = np.sum(bA, axis=1)
        k_in = np.sum(bA, axis=0)
        s_out = np.sum(A, axis=1)
        s_in = np.sum(A, axis=0)

        x0 = 0.1 * np.ones(16)
        args = (k_out, k_in, s_out, s_in)

        fun = lambda x: -mof.loglikelihood_prime_decm_exp(x, args)
        fun_jac = lambda x: -mof.loglikelihood_hessian_decm_exp(x, args)
        step_fun = lambda x: -mof.loglikelihood_decm_exp(x, args)
        lin_fun = lambda x: mof.linsearch_fun_DECM_exp(x, (mof.loglikelihood_decm_exp, args))
        hes_reg = sof.matrix_regulariser_function

        sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            fun_jac=fun_jac,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-1,
            max_steps=300,
            method="newton",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )
        sol = np.exp(-sol)

        ek = mof.expected_decm(sol)
        k = np.concatenate((k_out, k_in, s_out, s_in))
        err = np.max(np.abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 2: error = {}'.format(err))
        # print('method = {}, matrix {}x{} '.format('newton', 4, 4))

        # test result
        self.assertTrue(err < 1e-1)

    def test_newton_1(self):
        # x0 relies heavily on x0
        n, s = (4, 25)

        A = mg.random_weighted_matrix_generator_dense(
            n, sup_ext=10, sym=False, seed=s, intweights=True
        )
        A[0, :] = 0

        bA = np.array([[1 if aa != 0 else 0 for aa in a] for a in A])

        k_out = np.sum(bA, axis=1)
        k_in = np.sum(bA, axis=0)
        s_out = np.sum(A, axis=1)
        s_in = np.sum(A, axis=0)

        x0 = 0.1 * np.ones(16)
        args = (k_out, k_in, s_out, s_in)

        fun = lambda x: -mof.loglikelihood_prime_decm_exp(x, args)
        fun_jac = lambda x: -mof.loglikelihood_hessian_decm_exp(x, args)
        step_fun = lambda x: -mof.loglikelihood_decm_exp(x, args)
        lin_fun = lambda x: mof.linsearch_fun_DECM_exp(x, (mof.loglikelihood_decm_exp, args))
        hes_reg = sof.matrix_regulariser_function

        sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            fun_jac=fun_jac,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-1,
            max_steps=300,
            method="newton",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )
        sol = np.exp(-sol)

        ek = mof.expected_decm(sol)
        k = np.concatenate((k_out, k_in, s_out, s_in))
        err = np.max(np.abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 2: error = {}'.format(err))
        # print('method = {}, matrix {}x{} '.format('newton', 4, 4))

        # test result
        self.assertTrue(err < 1e-1)

    @unittest.skip("skip large graph")
    def test_newton_3(self):
        # convergence relies heavily on x0
        # n, s = (4, 35)
        n, s = (40, 35)
        A = mg.random_weighted_matrix_generator_dense(
            n, sup_ext=10, sym=False, seed=s, intweights=True
        )
        # A[0,:] = 0

        bA = np.array([[1 if aa != 0 else 0 for aa in a] for a in A])

        k_out = np.sum(bA, axis=1)
        k_in = np.sum(bA, axis=0)
        s_out = np.sum(A, axis=1)
        s_in = np.sum(A, axis=0)

        # x0 = 1*np.ones(4*n)
        x0 = np.concatenate((-1 * np.ones(2 * n), np.ones(2 * n)))
        args = (k_out, k_in, s_out, s_in)
        # x0[args == 0] = np.infty

        fun = lambda x: -mof.loglikelihood_prime_decm_exp(x, args)
        fun_jac = lambda x: -mof.loglikelihood_hessian_decm_exp(x, args)
        step_fun = lambda x: -mof.loglikelihood_decm_exp(x, args)
        lin_fun = lambda x: mof.linsearch_fun_DECM_exp(x, (step_fun,))
        hes_reg = sof.matrix_regulariser_function

        sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            fun_jac=fun_jac,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-2,
            max_steps=300,
            method="newton",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )
        sol = np.exp(-sol)

        ek = mof.expected_decm(sol)
        k = np.concatenate((k_out, k_in, s_out, s_in))
        err = np.max(np.abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 3: error = {}'.format(err))
        # print('method: {}, matrix {}x{}'.format('newton', n,n))

        # test result
        self.assertTrue(err < 1e-1)

    @unittest.skip("skip large graph")
    def test_newton_4(self):
        # convergence relies heavily on x0
        n, s = (40, 35)
        # n, s = (5, 35)
        A = mg.random_weighted_matrix_generator_dense(
            n, sup_ext=100, sym=False, seed=s, intweights=True
        )
        A[0, :] = 0

        bA = np.array([[1 if aa != 0 else 0 for aa in a] for a in A])

        k_out = np.sum(bA, axis=1)
        k_in = np.sum(bA, axis=0)
        s_out = np.sum(A, axis=1)
        s_in = np.sum(A, axis=0)

        x0 = 0.1 * np.ones(4 * n)
        # x0 = np.concatenate((-1*np.ones(2*n), np.ones(2*n)))
        args = (k_out, k_in, s_out, s_in)
        x0[np.concatenate(args) == 0] = 1e3

        fun = lambda x: -mof.loglikelihood_prime_decm_exp(x, args)
        fun_jac = lambda x: -mof.loglikelihood_hessian_decm_exp(x, args)
        step_fun = lambda x: -mof.loglikelihood_decm_exp(x, args)
        lin_fun = lambda x: mof.linsearch_fun_DECM_exp(x, (step_fun,))
        hes_reg = sof.matrix_regulariser_function

        sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            fun_jac=fun_jac,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-5,
            max_steps=100,
            method="newton",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )
        sol = np.exp(-sol)

        ek = mof.expected_decm(sol)
        k = np.concatenate((k_out, k_in, s_out, s_in))
        err = np.max(np.abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 4: error = {}'.format(err))
        # print('method: {}, matrix {}x{} with zeros'.format('newton', n,n))

        # test result
        self.assertTrue(err < 1e-1)

    def test_iterative_0(self):

        n, s = (4, 35)
        # n, s = (5, 35)
        A = mg.random_weighted_matrix_generator_dense(
            n, sup_ext=100, sym=False, seed=s, intweights=True
        )
        # A[0,:] = 0

        bA = np.array([[1 if aa != 0 else 0 for aa in a] for a in A])

        k_out = np.sum(bA, axis=1)
        k_in = np.sum(bA, axis=0)
        s_out = np.sum(A, axis=1)
        s_in = np.sum(A, axis=0)

        x0 = 0.1 * np.ones(n * 4)
        args = (k_out, k_in, s_out, s_in)
        x0[np.concatenate(args) == 0] = 1e3

        fun = lambda x: mof.iterative_decm_exp(x, args)
        step_fun = lambda x: -mof.loglikelihood_decm_exp(x, args)
        lin_fun = lambda x: mof.linsearch_fun_DECM_exp_fixed(x)
        hes_reg = sof.matrix_regulariser_function

        sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-10,
            max_steps=7000,
            method="fixed-point",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )

        sol = np.exp(-sol)
        ek = mof.expected_decm(sol)
        k = np.concatenate((k_out, k_in, s_out, s_in))
        err = np.max(np.abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 5: error = {}'.format(err))
        # print('method: {}, matrix {}x{} '.format('iterative', n,n))

        # test result
        self.assertTrue(err < 1)

    def test_iterative_1(self):

        n, s = (4, 35)
        # n, s = (5, 35)
        A = mg.random_weighted_matrix_generator_dense(
            n, sup_ext=100, sym=False, seed=s, intweights=True
        )
        A[0, :] = 0

        bA = np.array([[1 if aa != 0 else 0 for aa in a] for a in A])

        k_out = np.sum(bA, axis=1)
        k_in = np.sum(bA, axis=0)
        s_out = np.sum(A, axis=1)
        s_in = np.sum(A, axis=0)

        x0 = 0.1 * np.ones(n * 4)
        args = (k_out, k_in, s_out, s_in)
        x0[np.concatenate(args) == 0] = 1e3

        fun = lambda x: mof.iterative_decm_exp(x, args)
        step_fun = lambda x: -mof.loglikelihood_decm_exp(x, args)
        lin_fun = lambda x: mof.linsearch_fun_DECM_exp_fixed(x)
        hes_reg = sof.matrix_regulariser_function

        sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-10,
            max_steps=7000,
            method="fixed-point",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )

        sol = np.exp(-sol)
        ek = mof.expected_decm(sol)
        k = np.concatenate((k_out, k_in, s_out, s_in))
        err = np.max(np.abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 5: error = {}'.format(err))
        # print('method: {}, matrix {}x{} '.format('iterative', n,n))

        # test result
        self.assertTrue(err < 1)

    @unittest.skip("skip large graph")
    def test_iterative_2(self):

        n, s = (40, 35)
        # n, s = (5, 35)
        A = mg.random_weighted_matrix_generator_dense(
            n, sup_ext=100, sym=False, seed=s, intweights=True
        )
        # A[0,:] = 0

        bA = np.array([[1 if aa != 0 else 0 for aa in a] for a in A])

        k_out = np.sum(bA, axis=1)
        k_in = np.sum(bA, axis=0)
        s_out = np.sum(A, axis=1)
        s_in = np.sum(A, axis=0)

        x0 = 0.1 * np.ones(n * 4)
        args = (k_out, k_in, s_out, s_in)
        x0[np.concatenate(args) == 0] = 1e3

        fun = lambda x: mof.iterative_decm_exp(x, args)
        step_fun = lambda x: -mof.loglikelihood_decm_exp(x, args)
        lin_fun = lambda x: mof.linsearch_fun_DECM_exp(x, (step_fun,))
        hes_reg = sof.matrix_regulariser_function

        sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-10,
            max_steps=7000,
            method="fixed-point",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )

        sol = np.exp(-sol)
        ek = mof.expected_decm(sol)
        k = np.concatenate((k_out, k_in, s_out, s_in))
        err = np.max(np.abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 5: error = {}'.format(err))
        # print('method: {}, matrix {}x{} '.format('iterative', n,n))

        # test result
        self.assertTrue(err < 1)

    @unittest.skip("doesnt work")
    def test_iterative_3(self):

        n, s = (40, 35)
        # n, s = (5, 35)
        A = mg.random_weighted_matrix_generator_dense(
            n, sup_ext=100, sym=False, seed=s, intweights=True
        )
        A[0, :] = 0

        bA = np.array([[1 if aa != 0 else 0 for aa in a] for a in A])

        k_out = np.sum(bA, axis=1)
        k_in = np.sum(bA, axis=0)
        s_out = np.sum(A, axis=1)
        s_in = np.sum(A, axis=0)

        x0 = 0.1 * np.ones(n * 4)
        args = (k_out, k_in, s_out, s_in)
        x0[np.concatenate(args) == 0] = 1e3

        fun = lambda x: mof.iterative_decm_exp(x, args)
        step_fun = lambda x: -mof.loglikelihood_decm_exp(x, args)
        lin_fun = lambda x: mof.linsearch_fun_DECM_exp(x, (step_fun,))
        hes_reg = sof.matrix_regulariser_function

        sol = sof.solver(
            x0,
            fun=fun,
            step_fun=step_fun,
            linsearch_fun=lin_fun,
            tol=1e-6,
            eps=1e-10,
            max_steps=7000,
            method="fixed-point",
            verbose=False,
            regularise=True,
            full_return=False,
            linsearch=True,
            hessian_regulariser = hes_reg,
        )

        sol = np.exp(-sol)
        ek = mof.expected_decm(sol)
        k = np.concatenate((k_out, k_in, s_out, s_in))
        err = np.max(np.abs(ek - k))
        # debug
        # print(ek)
        # print(k)
        # print('\ntest 6: error = {}'.format(err))
        # print('method: {}, matrix {}x{} '.format('iterative', n,n))

        # test result
        self.assertTrue(err < 1)


if __name__ == "__main__":
    unittest.main()



import sys
import os

sys.path.append("../")
import NEMtropy.graph_classes as sample
import NEMtropy.graph_classes as sample_und
import NEMtropy.matrix_generator as mg
import numpy as np
import unittest


class MyTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_ECM_Dianati_random_dense_20_undir(self):

        network = mg.random_weighted_matrix_generator_dense(
            n=20, sup_ext=10, sym=True, seed=10, intweights=True
        )

        g = sample_und.UndirectedGraph(adjacency=network)

        g.solve_tool(
            model="ecm",
            method="fixed-point",
            max_steps=20000,
            verbose=False,
            initial_guess="random",
        )

        g._solution_error()

        # test result
        # print(g.error)

        self.assertTrue(g.error < 1)


if __name__ == "__main__":

    unittest.main()



import sys

sys.path.append("../")
import NEMtropy.graph_classes as sample
from NEMtropy.models_functions import *
import NEMtropy.matrix_generator as mg
import numpy as np
import unittest  # test tool
from scipy.optimize import approx_fprime


class MyTest(unittest.TestCase):
    def setUp(self):
        pass

    def test_iterative_dcm_exp(self):

        n, seed = (3, 42)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)

        # rd
        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "uniform"
        g.regularise = "identity"
        g._initialize_problem("dcm", "fixed-point")
        # theta = np.random.rand(6)
        theta = np.array([np.infty, 0.5, 0.5, 0.5, np.infty, 0.5])
        x0 = np.exp(-theta)

        f_sample = g.fun(x0)
        g.last_model = "dcm"
        g._set_solved_problem(f_sample)
        f_full = np.concatenate((g.x, g.y))
        f_full = -np.log(f_full)
        f_exp = iterative_dcm_exp(theta, g.args)
        g._set_solved_problem_dcm(f_exp)
        f_exp_full = np.concatenate((g.x, g.y))

        # f_exp_bis = iterative_dcm_exp_bis(theta, g.args)
        # print('normale ',f_exp)
        # print('bis',f_exp_bis)

        # debug
        # print(a)
        # print(theta, x0)
        # print(g.args)
        # print(f_full)
        # print(f_exp)
        # print(f_exp_full)

        # test result
        self.assertTrue(np.allclose(f_full, f_exp_full))

    def test_loglikelihood_dcm_exp(self):

        n, seed = (3, 42)
        # a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)
        a = np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0]])

        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "uniform"
        g.regularise = "identity"
        g._initialize_problem("dcm", "newton")
        theta = np.array([np.infty, 0.5, 0.5, 0.5, 0.5, 0.5])
        x0 = np.exp(-theta)

        k_out = g.args[0]
        k_in = g.args[1]
        nz_index_out = g.args[2]
        nz_index_in = g.args[3]

        f_sample = g.step_fun(x0)
        g.last_model = "dcm"
        f_exp = -loglikelihood_dcm_exp(theta, g.args)

        # debug
        # print(a)
        # print(theta, x0)
        # print(g.args)
        # print(f_sample)
        # print(f_sample)
        # print(f_exp)

        # test result
        self.assertTrue(f_sample == f_exp)

    def test_loglikelihood_prime_dcm_exp(self):

        n, seed = (3, 42)
        # a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)
        a = np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0]])

        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "uniform"
        g.regularise = "identity"
        g._initialize_problem("dcm", "newton")
        theta = np.array([0.5, 0.5, 0.5, 0.5, 0.5, 0.5])
        x0 = np.exp(-theta)

        k_out = g.args[0]
        k_in = g.args[1]
        nz_index_out = g.args[2]
        nz_index_in = g.args[3]

        f = lambda x: loglikelihood_dcm_exp(x, g.args)
        f_sample = approx_fprime(theta, f, epsilon=1e-6)
        g.last_model = "dcm"
        f_exp = loglikelihood_prime_dcm_exp(theta, g.args)

        # debug
        # print(a)
        # print(theta, x0)
        # print(g.args)
        # print(f_sample)
        # print(f_exp)

        # test result
        self.assertTrue(np.allclose(f_sample, f_exp))

    def test_loglikelihood_hessian_dcm_exp(self):

        n, seed = (3, 42)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)

        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "uniform"
        g.regularise = "identity"
        g._initialize_problem("dcm", "newton")
        theta = np.random.rand(2 * n)
        x0 = np.exp(-theta)

        k_out = g.args[0]
        k_in = g.args[1]
        nz_index_out = g.args[2]
        nz_index_in = g.args[3]

        f_sample = np.zeros((n * 2, n * 2))
        for i in range(n * 2):
            f = lambda x: loglikelihood_prime_dcm_exp(x, g.args)[i]
            f_sample[i, :] = approx_fprime(theta, f, epsilon=1e-6)

        f_exp = loglikelihood_hessian_dcm_exp(theta, g.args)

        # debug
        # print(a)
        # print(theta, x0)
        # print(g.args)
        # print(f_sample)
        # print(f_exp)

        # test result
        self.assertTrue(np.allclose(f_sample, f_exp))

    def test_loglikelihood_hessian_dcm_exp_emi(self):

        n, s = (50, 1)
        a = mg.random_binary_matrix_generator_custom_density(
            n=n, p=0.15, sym=False, seed=s
        )

        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "uniform"
        g.regularise = "identity"
        g._initialize_problem("dcm", "newton")

        k_out = g.args[0]
        k_in = g.args[1]
        nz_index_out = g.args[2]
        nz_index_in = g.args[3]

        n_rd = len(k_out)
        theta = np.random.rand(2 * n_rd)
        f_sample = np.zeros((n_rd * 2, n_rd * 2))
        for i in range(n_rd * 2):
            f = lambda x: loglikelihood_prime_dcm_exp(x, g.args)[i]
            f_sample[i, :] = approx_fprime(theta, f, epsilon=1e-6)

        f_exp = loglikelihood_hessian_dcm_exp(theta, g.args)

        # debug
        # print(a)
        # print(theta, x0)
        # print(g.args)
        # print(f_sample)
        # print(f_exp)

        # test result
        self.assertTrue(np.allclose(f_sample, f_exp))

    def test_loglikelihood_hessian_dcm_exp_emi_simmetry(self):

        n, s = (50, 1)
        # n,s =(2, 1)
        a = mg.random_binary_matrix_generator_custom_density(
            n=n, p=0.15, sym=False, seed=s
        )

        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "uniform"
        g.regularise = "identity"
        g._initialize_problem("dcm", "newton")

        k_out = g.args[0]
        k_in = g.args[1]
        nz_index_out = g.args[2]
        nz_index_in = g.args[3]

        n_rd = len(k_out)
        # print(n_rd/n)
        theta = np.random.rand(2 * n_rd)

        f_exp = loglikelihood_hessian_dcm_exp(theta, g.args)

        for i in range(2 * n_rd):
            for j in range(2 * n_rd):
                if f_exp[i, j] - f_exp[j, i] != 0:
                    print(i, j)
        # debug
        # print(a)
        # print(theta, x0)
        # print(g.args)
        # print(f_sample)
        # print(f_exp)

        # test result
        self.assertTrue(
            np.allclose(f_exp - f_exp.T, np.zeros((2 * n_rd, 2 * n_rd)))
        )

    def test_loglikelihood_hessian_dcm_exp_simmetry(self):

        n, seed = (3, 42)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)

        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "uniform"
        g.regularise = "identity"
        g._initialize_problem("dcm", "newton")
        theta = np.random.rand(2 * n)
        x0 = np.exp(-theta)

        k_out = g.args[0]
        k_in = g.args[1]
        nz_index_out = g.args[2]
        nz_index_in = g.args[3]

        f = loglikelihood_hessian_dcm_exp(theta, g.args)
        f_t = f.T

        # debug
        # print(a)
        # print(theta, x0)
        # print(g.args)
        # print(f-f_t)

        # test result
        self.assertTrue(np.allclose(f - f_t, np.zeros((2 * n, 2 * n))))

    def test_loglikelihood_hessian_diag_dcm_exp(self):

        n, seed = (3, 42)
        a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)

        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "uniform"
        g.regularise = "identity"
        g._initialize_problem("dcm", "newton")
        theta = np.random.rand(n * 2)
        x0 = np.exp(-theta)

        k_out = g.args[0]
        k_in = g.args[1]
        nz_index_out = g.args[2]
        nz_index_in = g.args[3]

        f_sample = np.zeros(2 * n)
        for i in range(2 * n):
            f = lambda x: loglikelihood_prime_dcm_exp(x, g.args)[i]
            f_sample[i] = approx_fprime(theta, f, epsilon=1e-6)[i]
        g.last_model = "dcm"
        f_exp = loglikelihood_hessian_diag_dcm_exp(theta, g.args)

        # debug
        # print(a)
        # print(theta, x0)
        # print(g.args)
        # print(f_sample)
        # print(f_exp)

        # test result
        self.assertTrue(np.allclose(f_sample, f_exp))

    def test_loglikelihood_hessian_diag_vs_normal_dcm_exp(self):

        n, seed = (3, 42)
        # a = mg.random_binary_matrix_generator_dense(n, sym=False, seed=seed)
        a = np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0]])

        g = sample.DirectedGraph(a)
        g.degree_reduction()
        g.initial_guess = "uniform"
        g.regularise = "identity"
        g._initialize_problem("dcm", "newton")
        theta = np.array([0.5, 0.5, 0.5, 0.5, 0.5, 0.5])
        x0 = np.exp(-theta)

        k_out = g.args[0]
        k_in = g.args[1]
        nz_index_out = g.args[2]
        nz_index_in = g.args[3]

        g.last_model = "dcm"
        f_diag = loglikelihood_hessian_diag_dcm_exp(theta, g.args)
        f_full = loglikelihood_hessian_dcm_exp(theta, g.args)
        f_full_d = np.diag(f_full)

        # debug
        # print(a)
        # print(theta, x0)
        # print(g.args)
        # print(f_sample)
        # print(f_exp)

        # test result
        self.assertTrue(np.allclose(f_diag, f_full_d))


if __name__ == "__main__":
    unittest.main()



"""This test checks the default setting for the seed is working.
When the seed is not manually set, it should be ranodmly chosen.
"""

import sys
import os
sys.path.append("../")
import NEMtropy.graph_classes as sample
import NEMtropy.matrix_generator as mg
import numpy as np
import unittest  # test tool
import random
import networkx as nx


class MyTest(unittest.TestCase):


    def setUp(self):
        pass


    def test_0(self):
        """test with 3 classes of cardinality 1
        and no zero degrees
        """
        """
        A = np.array(
            [
                [0, 1, 1, 0],
                [1, 0, 0, 1],
                [1, 0, 0, 0],
                [0, 1, 0, 0],
            ]
        )
        e = [(0,1), (0,2), (1,3)]
        d = [1,1,2,2]
        print(e)
        print(d)
        """
        N, seed = (10, 42)
        A = mg.random_binary_matrix_generator_dense(N, sym=True, seed=seed)
        # number of copies to generate

        g = sample.UndirectedGraph(A)

        g._solve_problem(
            model="cm",
            method="fixed-point",
            max_steps=100,
            verbose=False,
            linsearch=True,
            initial_guess="uniform",
        )

        x = g.x
        # g._solution_error()
        err = g.error

        # print('\ntest 5: error = {}'.format(g.error))
        n = 10
        output_dir = "sample/"
        # random.seed(100)
        g_list = []
        for i in range(n):
            g.ensemble_sampler(n=1, output_dir=output_dir)
            g_list.append(np.loadtxt("sample/0.txt"))

        appo = True
        old = g_list[0]
        for i in range(1, n):
            appo = appo*np.all(old == g_list[i])

        # debug
        """
        print('original dseq',d)
        print('original dseq sum ',g.dseq.sum())
        print('ensemble average dseq', d_emp)
        print('ensemble dseq sum ',np.array([d_emp[key] for key in d_emp.keys()]).sum())
        print(d_diff)
        print('empirical error', ensemble_error)
        print('theoretical error', err)
        """


        l = os.listdir(output_dir)
        for f in l:
            os.remove(output_dir + f)
        os.rmdir(output_dir)

        # test result
        self.assertTrue(not appo)


if __name__ == "__main__":
    unittest.main()



from setuptools import setup


with open('README.md', "r", encoding="utf-8") as fh:
    long_description = fh.read()


setup(
    name="NEMtropy",
    author="Nicolo' Vallarano, Emiliano Marchese, Matteo Bruno",
    author_email='nico.vallarano@gmail.com, emilianomarcheserc@gmail.com,'
                 ' matteobruno180@gmail.com',
    packages=["NEMtropy"],
    package_dir={'': 'src'},
    version="3.0.1",
    description="NEMtropy is a Maximum-Entropy toolbox for networks, it"
                " provides the user with a state of the art solver for a"
                " range variety of Maximum Entropy Networks models derived"
                " from the ERGM family. This module allows you to solve the"
                " desired model and generate a number of randomized graphs"
                " from the original one: the so-called graphs ensemble.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="GNU General Public License v3",
    url="https://github.com/nicoloval/NEMtropy/",
    download_url="https://github.com/nicoloval/NEMtropy/archive/refs/tags/v2.1.1.zip",
    keywords=['Network reconstruction', 'Networks Null Models',
              'Maximum Entropy Methods'],
    classifiers=[
                'License :: OSI Approved :: GNU Library or Lesser General'
                ' Public License (LGPL)',
                'Programming Language :: Python :: 3',
                'Programming Language :: Python :: 3.5',
                'Programming Language :: Python :: 3.6',
                'Programming Language :: Python :: 3.7',
                'Programming Language :: Python :: 3.8',
                ],
    install_requires=[
                      "numpy>=1.17",
                      "scipy>=1.4",
                      "networkx>=2.4",
                      "powerlaw>=1.4",
                      "tqdm>=4.5",
                      "bicm>=3.1"
                      ],
    extras_require={
        "dev": [
                "pytest>=6.0.1",
                "flake8>=3.8.3",
                "wheel>=0.35.1",
                "check-manifest>=0.44",
                "setuptools>=47.1.0",
                "twine>=3.2.0",
                "tox>=3.20.1",
                "powerlaw>=1.4.4",
                ],
        },
)



import numpy as np
import scipy.sparse
import scipy
import networkx as nx
import powerlaw as plw


def random_binary_matrix_generator_dense(n, sym=False, seed=None):
    """Generates a densely connected random binary adjacency matrix.

    :param n: Number of nodes.
    :type n: int
    :param sym: If True the adjacency matrix is simmetric, defaults to False.
    :type sym: bool, optional
    :param seed: Seed of the random process, defaults to None.
    :type seed: int, optional
    :return: Binary adjacency matrix.
    :rtype: numpy.ndarray
    """
    if not sym:
        np.random.seed(seed=seed)
        A = np.random.randint(0, 2, size=(n, n))
        # zeros on the diagonal
        np.fill_diagonal(A, 0)
        k_in = np.sum(A, axis=0)
        k_out = np.sum(A, axis=1)
        for ind, ki, ko in zip(np.arange(k_in.shape[0]), k_in, k_out):
            if (ki == 0) and (ko == 0):
                while (np.sum(A[:, ind]) == 0) and (np.sum(A[ind, :]) == 0):
                    if np.random.random() > 0.5:
                        A[np.random.randint(A.shape[0]), ind] = 1
                    else:
                        A[ind, np.random.randint(A.shape[0])] = 1

                    A[ind, ind] = 0
        return A
    else:
        np.random.seed(seed=seed)
        A = np.random.randint(0, 2, size=(n, n))
        A = ((A + A.T) / 2).astype(int)
        # zeros on the diagonal
        np.fill_diagonal(A, 0)
        k = np.sum(A, axis=0)
        for ind, kk in zip(np.arange(k.shape[0]), k):
            if kk == 0:
                while np.sum(A[:, ind]) == 0:
                    indices = np.random.randint(A.shape[0])
                    A[indices, ind] = 1
                    A[ind, indices] = 1
                    A[ind, ind] = 0
        return A


def random_binary_matrix_generator_custom_density(
    n, p=0.1, sym=False, seed=None
):
    """Generates a random binary adjacency matrix with density determined by the parameter p.

    :param n: Number of nodes.
    :type n: int
    :param p: Probability of observing a link between a couple of nodes,
     defaults to 0.1.
    :type p: float, optional
    :param sym: If True the adjacency matrix is simmetric, defaults to False.
    :type sym: bool, optional
    :param seed: Seed of the random process, defaults to None.
    :type seed: int, optional
    :return: Binary adjacency matrix.
    :rtype: numpy.ndarray
    """
    if not sym:
        np.random.seed(seed=seed)
        A = np.zeros(shape=(n, n))
        for i in range(n):
            for j in range(n):
                if np.random.random() <= p:
                    A[i, j] = 1
        np.fill_diagonal(A, 0)
        k_in = np.sum(A, axis=0)
        k_out = np.sum(A, axis=1)
        for ind, ki, ko in zip(np.arange(k_in.shape[0]), k_in, k_out):
            if (ki == 0) and (ko == 0):
                while (np.sum(A[:, ind]) == 0) and (np.sum(A[ind, :]) == 0):
                    if np.random.random() > 0.5:
                        A[np.random.randint(A.shape[0]), ind] = 1
                    else:
                        A[ind, np.random.randint(A.shape[0])] = 1

                    A[ind, ind] = 0
        return A
    else:
        np.random.seed(seed=seed)
        A = np.zeros(shape=(n, n))
        for i in range(n):
            for j in range(i + 1, n):
                if np.random.random() <= p:
                    A[i, j] = 1
                    A[j, i] = A[i, j]
        np.fill_diagonal(A, 0)
        degree = np.sum(A, axis=0)
        for ind, k in enumerate(degree):
            if k == 0:
                while np.sum(A[:, ind]) == 0:
                    indices = np.random.randint(A.shape[0])
                    if indices != ind:
                        A[ind, indices] = 1
                        A[indices, ind] = A[ind, indices]
                    A[ind, ind] = 0
        return A


def random_weighted_matrix_generator_dense(
    n, sup_ext=10, sym=False, seed=None, intweights=False
):
    """Generates a densely connected random weighted adjacency matrix.

    :param n: Number of nodes.
    :type n: int
    :param sup_ext: Maximum strength, defaults to 10.
    :type sup_ext: int, float, optional
    :param sym: If True the adjacency matrix is simmetric, defaults to False.
    :type sym: bool, optional
    :param seed: Seed of the random process, defaults to None.
    :type seed: int, optional
    :param intweights: If True weights are integers, defaults to False.
    :type intweights: bool, optional
    :return: Weighted adjacency matrix.
    :rtype: numpy.ndarray
    """
    if not sym:
        np.random.seed(seed=seed)
        A = np.random.random(size=(n, n)) * sup_ext
        np.fill_diagonal(A, 0)

        k_in = np.sum(A, axis=0)
        k_out = np.sum(A, axis=1)
        for ind, ki, ko in zip(np.arange(k_in.shape[0]), k_in, k_out):
            if (ki == 0) and (ko == 0):
                while (np.sum(A[:, ind]) == 0) and (np.sum(A[ind, :]) == 0):
                    if np.random.random() > 0.5:
                        A[np.random.randint(A.shape[0]), ind] = (
                            np.random.random() * sup_ext
                        )
                    else:
                        A[ind, np.random.randint(A.shape[0])] = (
                            np.random.random() * sup_ext
                        )

                    A[ind, ind] = 0
        if intweights:
            return np.ceil(A)
        else:
            return A
    else:
        np.random.seed(seed=seed)
        b = np.random.random(size=(n, n)) * sup_ext
        A = (b + b.T) / 2
        np.fill_diagonal(A, 0)

        degree = np.sum(A, axis=0)
        for ind, k in enumerate(degree):
            if k == 0:
                while np.sum(A[:, ind]) == 0:
                    indices = np.random.randint(A.shape[0])
                    if indices != ind:
                        A[ind, indices] = np.random.random() * sup_ext
                        A[indices, ind] = A[ind, indices]
        if intweights:
            return np.ceil(A)
        else:
            return A


def random_weighted_matrix_generator_uniform_custom_density(
    n, p=0.1, sup_ext=10, sym=False, seed=None, intweights=False
):
    """Generates a random weighted adjacency matrix with density determined
    by the parameter p and weights extracted from a uniform distribution.

    :param n: Number of nodes.
    :type n: int
    :param p: Probability of observing a link between a couple of nodes,
     defaults to 0.1.
    :type p: float, optional
    :param sup_ext: Maximum strength, defaults to 10.
    :type sup_ext: int or float, optional
    :param sym: If True the adjacency matrix is simmetric, defaults to False.
    :type sym: bool, optional
    :param seed: Seed of the random process, defaults to None.
    :type seed: int, optional
    :param intweights: If True weights are integers, defaults to False.
    :type intweights: bool, optional
    :return: Weighted adjacency matrix.
    :rtype: numpy.ndarray
    """
    if not sym:
        np.random.seed(seed=seed)
        A = np.zeros(shape=(n, n))
        for i in range(n):
            for j in range(n):
                if np.random.random() <= p:
                    A[i, j] = np.random.random() * sup_ext
        np.fill_diagonal(A, 0)
        k_in = np.sum(A, axis=0)
        k_out = np.sum(A, axis=1)
        for ind, ki, ko in zip(np.arange(k_in.shape[0]), k_in, k_out):
            if (ki == 0) and (ko == 0):
                while (np.sum(A[:, ind]) == 0) and (np.sum(A[ind, :]) == 0):
                    if np.random.random() > 0.5:
                        A[np.random.randint(A.shape[0]), ind] = (
                            np.random.random() * sup_ext
                        )
                    else:
                        A[ind, np.random.randint(A.shape[0])] = (
                            np.random.random() * sup_ext
                        )

                    A[ind, ind] = 0
        if intweights:
            return np.ceil(A)
        else:
            return A
    else:
        np.random.seed(seed=seed)
        A = np.zeros(shape=(n, n))
        for i in range(n):
            for j in range(i + 1, n):
                if np.random.random() <= p:
                    A[i, j] = np.random.random() * sup_ext
                    A[j, i] = A[i, j]
        degree = np.sum(A, axis=0)
        for ind, k in enumerate(degree):
            if k == 0:
                while np.sum(A[:, ind]) == 0:
                    indices = np.random.randint(A.shape[0])
                    if indices != ind:
                        A[ind, indices] = np.random.random() * sup_ext
                        A[indices, ind] = A[ind, indices]
        if intweights:
            return np.ceil(A)
        else:
            return A


def random_weighted_matrix_generator_gaussian_custom_density(
    n, mean, sigma, p=0.1, sym=False, seed=None, intweights=False
):
    """Generates a random weighted adjacency matrix with density determined
    by the parameter p and weights extracted from a gaussian distribution.

    :param n: Number of nodes.
    :type n: int
    :param mean: Mean of the gaussian.
    :type mean: int, float
    :param sigma: Sigma of the gaussian.
    :type sigma: int, float
    :param p: Probability of observing a link between a couple of nodes,
     defaults to 0.1.
    :type p: float, optional
    :param sym: If True the adjacency matrix is simmetric, defaults to False.
    :type sym: bool, optional
    :param seed: Seed of the random process, defaults to None.
    :type seed: int, optional
    :param intweights: If True weights are integers, defaults to False.
    :type intweights: bool, optional
    :return: Weighted adjacency matrix.
    :rtype: numpy.ndarray
    """
    if not sym:
        np.random.seed(seed=seed)
        A = np.zeros(shape=(n, n))
        for i in range(n):
            for j in range(n):
                if np.random.random() <= p:
                    A[i, j] = np.random.normal(loc=mean, scale=sigma)
        np.fill_diagonal(A, 0)
        k_in = np.sum(A, axis=0)
        k_out = np.sum(A, axis=1)
        for ind, ki, ko in zip(np.arange(k_in.shape[0]), k_in, k_out):
            if (ki == 0) and (ko == 0):
                while (np.sum(A[:, ind]) == 0) and (np.sum(A[ind, :]) == 0):
                    if np.random.random() > 0.5:
                        A[
                            np.random.randint(A.shape[0]), ind
                        ] = np.random.normal(loc=mean, scale=sigma)
                    else:
                        A[
                            ind, np.random.randint(A.shape[0])
                        ] = np.random.normal(loc=mean, scale=sigma)

                    A[ind, ind] = 0
        if intweights:
            return np.ceil(A)
        else:
            return A
    else:
        np.random.seed(seed=seed)
        A = np.zeros(shape=(n, n))
        for i in range(n):
            for j in range(i + 1, n):
                if np.random.random() <= p:
                    A[i, j] = np.random.normal(loc=mean, scale=sigma)
                    A[j, i] = A[i, j]
        np.fill_diagonal(A, 0)
        degree = np.sum(A, axis=0)
        for ind, k in enumerate(degree):
            if k == 0:
                while np.sum(A[:, ind]) == 0:
                    indices = np.random.randint(A.shape[0])
                    if indices != ind:
                        A[indices, ind] = np.random.normal(
                            loc=mean, scale=sigma
                        )
                        A[ind, indices] = A[indices, ind]
        if intweights:
            return np.ceil(A)
        else:
            return A


def random_binary_matrix_generator_custom_density_sparse(
    n, p=0.1, sym=False, seed=None
):
    """Generates a random binary adjacency sparse matrix
    with density determined by the parameter p.

    :param n: Number of nodes.
    :type n: int
    :param p: Probability of observing a link between a couple of nodes,
     defaults to 0.1.
    :type p: float, optional
    :param sym: If True the adjacency matrix is simmetric, defaults to False.
    :type sym: bool, optional
    :param seed: seed of the random process, defaults to None.
    :type seed: int, optional
    :return: a binary sparse adjacency matrix.
    :rtype: scipy.sparse_matrix
    """
    if sym:
        nx_graph = nx.fast_gnp_random_graph(n, p=p, seed=seed)
        largest_cc = max(nx.connected_components(nx_graph), key=len)
        nx_graph_lcc = nx_graph.subgraph(largest_cc).copy()
        adj_sparse = nx.to_scipy_sparse_matrix(nx_graph_lcc)
        return adj_sparse

    else:
        nx_graph = nx.fast_gnp_random_graph(n, p=p, seed=seed, directed=True)
        largest_cc = max(nx.weakly_connected_components(nx_graph), key=len)
        nx_graph_lcc = nx_graph.subgraph(largest_cc).copy()
        adj_sparse = nx.to_scipy_sparse_matrix(nx_graph_lcc)
        return adj_sparse


def random_uniform_weighted_matrix_generator_custom_density_sparse(
    n, sup_ext, p=0.1, sym=False, seed=None, intweights=False
):
    """Generates a random weighted adjacency sparse matrix with density
    determined by the parameter p and weights extracted
    from a uniform distribution.

    :param n: Number of nodes.
    :type n: int
    :param sup_ext: Maximum strength, defaults to 10.
    :type sup_ext: int, float, optional
    :param p: Probability of observing a link between a couple of nodes,
     defaults to 0.1.
    :type p: float, optional
    :param sym: If True the adjacency matrix is simmetrix, defaults to False.
    :type sym: bool, optional
    :param seed: Seed of the random process, defaults to None.
    :type seed: int, optional
    :param intweights: If True weights are integers, defaults to False.
    :type intweights: bool, optional
    :return: Weighted sparse adjacency matrix.
    :rtype: scipy.sparse_matrix
    """
    if sym:
        nx_graph = nx.fast_gnp_random_graph(n, p=p, seed=seed)
        largest_cc = max(nx.connected_components(nx_graph), key=len)
        nx_graph_lcc = nx_graph.subgraph(largest_cc).copy()
        adj_sparse = nx.to_scipy_sparse_matrix(nx_graph_lcc)
        adj_sparse = adj_sparse.astype(np.float64)

        # Weigths
        raw_ind, col_ind = scipy.sparse.triu(adj_sparse).nonzero()
        if intweights:
            for i, j in zip(raw_ind, col_ind):
                adj_sparse[i, j] = np.ceil(np.random.random() * sup_ext)
                adj_sparse[j, i] = adj_sparse[i, j]
        else:
            for i, j in zip(raw_ind, col_ind):
                adj_sparse[i, j] = np.random.random() * sup_ext
                adj_sparse[j, i] = adj_sparse[i, j]

        return adj_sparse

    else:
        nx_graph = nx.fast_gnp_random_graph(n, p=p, seed=seed, directed=True)
        largest_cc = max(nx.weakly_connected_components(nx_graph), key=len)
        nx_graph_lcc = nx_graph.subgraph(largest_cc).copy()
        adj_sparse = nx.to_scipy_sparse_matrix(nx_graph_lcc)
        adj_sparse = adj_sparse.astype(np.float64)

        # Weigths
        raw_ind, col_ind = adj_sparse.nonzero()
        if intweights:
            for i, j in zip(raw_ind, col_ind):
                adj_sparse[i, j] = np.ceil(np.random.random() * sup_ext)
        else:
            for i, j in zip(raw_ind, col_ind):
                adj_sparse[i, j] = np.random.random() * sup_ext

        return adj_sparse


def random_gaussian_weighted_matrix_generator_custom_density_sparse(
    n, mean, sigma, p=0.1, sym=False, seed=None, intweights=False
):
    """Generates a random weighted adjacency sparse matrix with density
    determined by the parameter p and weights
    extracted from a gaussian distribution.

    :param n: Number of nodes.
    :type n: int
    :param mean: Mean of the gaussian.
    :type mean: int, float
    :param sigma: Sigma of the gaussian.
    :type sigma: int, float
    :param p: Probability of observing a link between a couple of nodes,
     defaults to 0.1.
    :type p: float, optional
    :param sym: If True the adjacency matrix is simmetrix, defaults to False.
    :type sym: bool, optional
    :param seed: Seed of the random process, defaults to None.
    :type seed: int, optional
    :param intweights: If True weights are integers, defaults to False.
    :type intweights: bool, optional
    :return: Weighted sparse adjacency matrix.
    :rtype: scipy.sparse_matrix
    """
    if sym:
        nx_graph = nx.fast_gnp_random_graph(n, p=p, seed=seed)
        largest_cc = max(nx.connected_components(nx_graph), key=len)
        nx_graph_lcc = nx_graph.subgraph(largest_cc).copy()
        adj_sparse = nx.to_scipy_sparse_matrix(nx_graph_lcc)
        adj_sparse = adj_sparse.astype(np.float64)

        # Weigths
        raw_ind, col_ind = scipy.sparse.triu(adj_sparse).nonzero()
        if intweights:
            for i, j in zip(raw_ind, col_ind):
                adj_sparse[i, j] = np.ceil(
                    np.random.normal(loc=mean, scale=sigma)
                )
                adj_sparse[j, i] = adj_sparse[i, j]
        else:
            for i, j in zip(raw_ind, col_ind):
                adj_sparse[i, j] = np.random.normal(loc=mean, scale=sigma)
                adj_sparse[j, i] = adj_sparse[i, j]

        return adj_sparse

    else:
        nx_graph = nx.fast_gnp_random_graph(n, p=p, seed=seed, directed=True)
        largest_cc = max(nx.weakly_connected_components(nx_graph), key=len)
        nx_graph_lcc = nx_graph.subgraph(largest_cc).copy()
        adj_sparse = nx.to_scipy_sparse_matrix(nx_graph_lcc)
        adj_sparse = adj_sparse.astype(np.float64)

        # Weigths
        raw_ind, col_ind = adj_sparse.nonzero()
        if intweights:
            for i, j in zip(raw_ind, col_ind):
                adj_sparse[i, j] = np.ceil(
                    np.random.normal(loc=mean, scale=sigma)
                )
        else:
            for i, j in zip(raw_ind, col_ind):
                adj_sparse[i, j] = np.random.normal(loc=mean, scale=sigma)

        return adj_sparse


def random_graph_nx(
    n, p, sup_ext, alpha, seed=None, is_weighted=None, is_sparse=False
):
    """Retunrs an undirected weighted adjacency matrix using erdos-renyi model
    to generate the binary adjacency and assigning weights extracted'
    from a uniform, gaussian or powerlaw distribution.

    :param n: umber of nodes.
    :type n: int
    :param p: Probability of observing a link between a couple of nodes,
     defaults to 0.1.
    :type p: float, optional
    :param sup_ext: Maximum strength, defaults to 10.
    :type sup_ext: int, float, optional
    :param alpha: Powerlaw exponent.
    :type alpha: float
    :param seed: Seed of the random process, defaults to None.
    :type seed: int, optional
    :param is_weighted: If True the adjacency matrix is weighted,
     defaults to None.
    :type is_weighted: bool, optional
    :param is_sparse: If True the adjacency matrix is sparse,
     defaults to False.
    :type is_sparse: bool, optional
    :return: Adjacency matrix.
    :rtype: numpy.ndarray, scipy.sparse_matrix
    """
    if seed is None:
        seed = np.random.randint(0, n ** 2)
    nx_graph = nx.fast_gnp_random_graph(n=n, p=p, seed=seed)
    largest_cc = max(nx.connected_components(nx_graph), key=len)
    nx_graph_lcc = nx_graph.subgraph(largest_cc).copy()

    np.random.seed(seed=seed)
    if is_weighted == "uniform":
        for e in nx_graph_lcc.edges:
            nx_graph_lcc[e[0]][e[1]]["weight"] = np.random.randint(0, sup_ext)
    elif is_weighted == "gaussian":
        for e in nx_graph_lcc.edges:
            nx_graph_lcc[e[0]][e[1]]["weight"] = np.random.normal(
                loc=sup_ext, scale=sup_ext / 5.5
            )
    elif is_weighted == "powerlaw":
        for e in nx_graph_lcc.edges:
            nx_graph_lcc[e[0]][e[1]]["weight"] = plw.Power_Law(
                xmin=1, xmax=sup_ext, parameters=[alpha], discrete=True
            ).generate_random(1)

    if is_sparse:
        adjacency = nx.to_scipy_sparse_matrix(nx_graph_lcc)
    else:
        adjacency = nx.to_numpy_array(nx_graph_lcc)

    return adjacency


def barabasi_albert_graph_nx(
    n, m, sup_ext, alpha, seed=None, is_weighted=None, is_sparse=False
):
    """Generates a undirected weighted adjacency matrix using barabasi-albert model
    for the binary part and assigning weights extracted from
    a uniform, gaussian or powerlaw distribution.

    :param n: Number of nodes.
    :type n: int
    :param m: Number of edges to attach from a new node to existing nodes.
    :type m: int
    :param sup_ext: Maximum strength, defaults to 10.
    :type sup_ext: int, float, optional
    :param alpha: Powerlaw exponent.
    :type alpha: float
    :param seed: Seed of the random process, defaults to None.
    :type seed: int, optional
    :param is_weighted: If True the adjacency matrix is weighted,
     defaults to None.
    :type is_weighted: bool, optional
    :param is_sparse: If True the adjacency matrix is sparse,
     defaults to False.
    :type is_sparse: bool, optional
    :return: Adjacency matrix.
    :rtype: numpy.ndarray, scipy.sparse_matrix
    """
    if seed is None:
        seed = np.random.randint(0, n ** 2)
    nx_graph = nx.barabasi_albert_graph(n, m, seed=seed)
    largest_cc = max(nx.connected_components(nx_graph), key=len)
    nx_graph_lcc = nx_graph.subgraph(largest_cc).copy()

    np.random.seed(seed=seed)
    if is_weighted == "uniform":
        for e in nx_graph_lcc.edges:
            nx_graph_lcc[e[0]][e[1]]["weight"] = np.random.randint(0, sup_ext)
    elif is_weighted == "gaussian":
        for e in nx_graph_lcc.edges:
            nx_graph_lcc[e[0]][e[1]]["weight"] = np.random.normal(
                loc=sup_ext, scale=sup_ext / 5.5
            )
    elif is_weighted == "powerlaw":
        for e in nx_graph_lcc.edges:
            nx_graph_lcc[e[0]][e[1]]["weight"] = plw.Power_Law(
                xmin=1, xmax=sup_ext, parameters=[alpha], discrete=True
            ).generate_random(1)

    if is_sparse:
        adjacency = nx.to_scipy_sparse_matrix(nx_graph_lcc)
    else:
        adjacency = nx.to_numpy_array(nx_graph_lcc)

    return adjacency
